--[[ ---------------------------------------------------------------------------

DebuffEnough: Debuff monitor

DebuffEnough is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DebuffEnough is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with DebuffEnough.	If not, see <http://www.gnu.org/licenses/>.

----------------------------------------------------------------------------- ]]

--[[ ---------------------------------------------------------------------------
	 Addon and libraries
----------------------------------------------------------------------------- ]]
DebuffEnough = nil

if LibStub("LibLogger-1.0", true) then
    DebuffEnough = LibStub("AceAddon-3.0"):NewAddon("DebuffEnough", "AceEvent-3.0", "AceBucket-3.0", "AceTimer-3.0", "AceConsole-3.0", "LibLogger-1.0")
else
	DebuffEnough = LibStub("AceAddon-3.0"):NewAddon("DebuffEnough", "AceEvent-3.0", "AceBucket-3.0", "AceTimer-3.0", "AceConsole-3.0")
end

local L = LibStub("AceLocale-3.0"):GetLocale("DebuffEnough")
local LDB = LibStub:GetLibrary("LibDataBroker-1.1")
local TQ = LibStub("LibTalentQuery-1.0")


--[[ ---------------------------------------------------------------------------
	 Ace3 initialization
----------------------------------------------------------------------------- ]]
function DebuffEnough:OnInitialize()

	-- Initialize persistent addon variables
	self.isDebuffEnough = true
	self.isDebuffWarning = false
	self.talents = {}
	self.pets = {}
	self.trackedDebuffs = {}
	self.haveTalentsFor = 0
	self.groupSize = 0
	
	self.Anchor = nil
	self.Display = nil
	self.Grip = nil
	self.tooltip = ""
	self.optionsFrame = nil
	self.isShowingTooltip = false
	self.timeSinceDebuffEnough = 0
		
	-- Set up saved variables and config options
	self.db = LibStub("AceDB-3.0"):New("DebuffEnoughDB", self.defaults, "Default")

	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileDeleted","OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	
	self.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("DebuffEnough", self.options, {"debuffenough", "de"})
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("DebuffEnough", "DebuffEnough")
	
	-- Logging
	if LibStub("LibLogger-1.0", true) then
		self:SetLogLevel(self.db.profile.logLevel)
		self:SetPerformanceMode(true)
	end
	
	self.dobj = LDB:NewDataObject("DebuffEnough", {
		icon = "Interface\\Icons\\Ability_Creature_Cursed_02",
		label = "DebuffEnough",
    	text = L["Yes"],
    	OnClick = function(frame, button)
            if button == "LeftButton" then
            	if not self:GetProfileParam("enable") then
            		DebuffEnough:ToggleConfigDialog()
            	else
                	DebuffEnough:ScanRaid()
                	DebuffEnough:ScanDebuffs()
                end
            elseif button == "RightButton" then
                --DebuffEnough:PrintResults("player")
                --DebuffEnough:PrintResults("pet")
            end
        end,
    	OnTooltipShow = function(tooltip)
    		local txt = nil
			if not self:GetProfileParam("enable") then
				txt = "DebuffEnough".."\n"..L["Currently disabled"]
			else
				local talentScan = "\n|cffffffffTalent Scans: "..self.haveTalentsFor.."/"..self.groupSize.."|r"
				txt = "DebuffEnough"..talentScan..self.tooltip..L["Hint"]
			end
			
			if tooltip and tooltip.AddLine then
				tooltip:AddLine(txt)
			end
		end
	})
	
	LDB.RegisterCallback(self, "LibDataBroker_AttributeChanged_DebuffEnough_text", "UpdateDisplay")
	LDB.RegisterCallback(self, "LibDataBroker_AttributeChanged_DebuffEnough_debuffenoughtooltiptext", "UpdateDisplay")
	
	TQ.RegisterCallback(self, "TalentQuery_Ready")
	
	-- Set up display
	self:CreateFrame()
	
	-- Register the events that might enable the addon
	self:RegisterBucketEvent("RAID_ROSTER_UPDATE", .2, "CheckRaidChange")
	self:RegisterBucketEvent("PARTY_MEMBERS_CHANGED", .2, "CheckRaidChange")
	
end


--[[ ---------------------------------------------------------------------------
	 Ace3 enable addon
----------------------------------------------------------------------------- ]]
function DebuffEnough:OnEnable()

	local isSolo = GetNumRaidMembers() == 0 and GetNumPartyMembers() == 0
	
	if isSolo and not self:GetProfileParam("solo") then
		self:SetProfileParam("enable", false)
	end

	if self:GetProfileParam("enable") then
		self:DoEnable()
	end

end


--[[ ---------------------------------------------------------------------------
	 Enables addon 
----------------------------------------------------------------------------- ]]
function DebuffEnough:DoEnable()

	if self.debug then self:debug("Addon enabled") end
	
	self:SetProfileParam("enable", true) 
	
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "ScanDebuffs")
	self:RegisterEvent("UNIT_AURA")
	
	self:SetAnchors(true)
	self:UpdateVisible()
	self:ScanRaid()

end


--[[ ---------------------------------------------------------------------------
	 Ace3 disable addon
----------------------------------------------------------------------------- ]]
function DebuffEnough:OnDisable()

	self:DoDisable()

end


--[[ ---------------------------------------------------------------------------
	 Disables addon
----------------------------------------------------------------------------- ]]
function DebuffEnough:DoDisable()

	if self.debug then self:debug("Addon disabled") end
	
	self:SetProfileParam("enable", false) 
	
	self:UnregisterAllEvents()

	self:UpdateVisible()
	
end


--[[ ---------------------------------------------------------------------------
	 Check to see if we need to activate the addon or run a check based on
	 changes to the party/raid
----------------------------------------------------------------------------- ]]
function DebuffEnough:CheckRaidChange()

	local isSolo = GetNumRaidMembers() == 0 and GetNumPartyMembers() == 0
	if self.debug then self:debug("Checking raid change, solo: %s", tostring(isSolo)) end

	if not isSolo and not self:GetProfileParam("enable") then
		self:DoEnable()
	elseif isSolo and self:GetProfileParam("solo") and not self:GetProfileParam("enable") then
		self:DoEnable()
	elseif isSolo and self:GetProfileParam("enable") and not self:GetProfileParam("solo") then
		self:DoDisable()
	end
	
	self:ScanRaid()
	
end


--[[ ---------------------------------------------------------------------------
	 Scans the makeup of the raid to determine what debuffs are available
----------------------------------------------------------------------------- ]]
function DebuffEnough:ScanRaid()

	if self.debug then self:debug("Scanning raid/party") end
	
	local newTalents = {DEATHKNIGHT = {}, DRUID = {}, HUNTER = {}, MAGE = {}, PALADIN = {}, PRIEST = {}, ROGUE = {}, SHAMAN = {}, WARLOCK = {}, WARRIOR = {}}
	local newPets = {}
	self.haveTalentsFor = 0
	self.groupSize = 1
		
	-- set up to iterate over raid/party
	local groupType = "party"
	local groupSize = GetNumPartyMembers()

	if GetNumRaidMembers() > 0 then
		groupType = "raid"
		groupSize = GetNumRaidMembers()
	end

	if groupSize > 1 then self.groupSize = groupSize end
	
	local i = 0
	local unit = "player"
		
	-- get talents for each party/raid member
	while i <= groupSize do
	
		if UnitExists(unit) then
			
			local name = UnitName(unit)
			local class = select(2, UnitClass(unit))
		
			if name and class then
				if (self.talents[class] and self.talents[class][name])then
					newTalents[class][name] = self.talents[class][name]
				else
					newTalents[class][name] = {{}, {}, {}}
				end
			end
			
			if UnitExists(unit.."pet") then
				newPets[name] = UnitCreatureFamily(unit.."pet")
			end
			
		end
	
		i = i + 1
		unit = groupType..i
	
	end
	
	self.talents = newTalents
	self.pets = newPets
	self.trackedDebuffs = {}

	-- check for missing talents
	for className, classes in pairs(self.talents) do
		for name, talents in pairs(classes) do
			if not next(talents[1]) then
				self:GetTalents(name)
			else
				self.haveTalentsFor = self.haveTalentsFor + 1
			end
		end
	end
		
	-- figure out what debuffs we have to work with
	for spellName, spellData in pairs(DebuffEnough.debuffs) do
		for name, talents in pairs(self.talents[spellData.class]) do
			
			local debuffValue = self:GetDebuffValue(name, talents, spellName, spellData)
				
			if debuffValue > 0 then
					
				if not self.trackedDebuffs[spellName] then
					self.trackedDebuffs[spellName] = {names = {}, slot = spellData.slot, applied = false}
				end
					
				self.trackedDebuffs[spellName].names[name] = debuffValue
				
			end
			
		end
	end

end


--[[ ---------------------------------------------------------------------------
	 Queue up a talent query for the given unit (or do it right away if its
	 for the player)
----------------------------------------------------------------------------- ]]
function DebuffEnough:GetTalents(name)

	if self.debug then self:debug("Queuing up getting talents for %s", name) end
	
	if (UnitIsConnected(name) and UnitIsVisible(name) and CanInspect(name)) then
	
		if UnitIsUnit(name, "player") then
			self:TalentQuery_Ready(_, name)
		else
			TQ:Query(name)
		end
	
	end

end


--[[ ---------------------------------------------------------------------------
	 Callback function for unit talent inspection, called from LibTalentQuery
	 after we call Query() and the ready event fires
----------------------------------------------------------------------------- ]]
function DebuffEnough:TalentQuery_Ready(_, name)

	local isNotPlayer = not UnitIsUnit(name, "player")
	local class = select(2, UnitClass(name))
	
	if not self.talents[class] or not self.talents[class][name] then return end
	
	if self.debug then self:debug("Talent query ready for %s (isNotPlayer=%s)", name, tostring(isNotPlayer)) end
	
	for i = 1, 3 do
		for j = 1, GetNumTalents(i, isNotPlayer) do
			self.talents[class][name][i][j] = select(5, GetTalentInfo(i, j, isNotPlayer))
		end
	end

	self:ScanRaid()

end


--[[ ---------------------------------------------------------------------------
	 Get the computed value of the debuff for the given talent spec, returns 0
	 if debuff requires talent and unit doesn't have it
----------------------------------------------------------------------------- ]]
function DebuffEnough:GetDebuffValue(name, talents, spellName, spellData)

	if self.trace then self:trace("Get debuff value for %s for %s", name, spellName) end

	local debuffValue = 0
	local hasRequiredTalent = false
	
	-- get value from required talent if applicable
	if spellData.requiresTalentTab then
	
		local talentValue = talents[spellData.requiresTalentTab][spellData.requiresTalentNum]
		if not talentValue then 
			talentValue = 0
			self:GetTalents(name)
		end
		
		debuffValue = talentValue * spellData.amountPerTalentPoint
		
		if debuffValue > 0 then
	
			hasRequiredTalent = true
		end
		
	end
	
	-- get value from normal class ability for level (can apply to pets as well)
	if (spellData.levelForNewRank and
		(not spellData.creatureType or 
		(hasRequiredTalent and spellData.creatureType == self.pets[name])))
	then
		for index, level in ipairs(spellData.levelForNewRank) do
			if (UnitLevel(name) >= level) then
				debuffValue = spellData.amountAtRank[index]
			end
		end
	end
	
	-- get value for improved talent if applicable
	if spellData.improvedByTalentTab then
	
		local talentValue = talents[spellData.improvedByTalentTab][spellData.improvedByTalentNum]
		if not talentValue then
			talentValue = 0
			self:GetTalents(name)
		end
		
		if spellData.amountPerTalentPoint then
			debuffValue = debuffValue + (spellData.amountPerTalentPoint * talentValue)
		elseif spellData.percentagePerTalentPoint then
			debuffValue = debuffValue + (debuffValue * ((spellData.percentagePerTalentPoint * talentValue) / 100))
		end
		
	end
	
	return debuffValue

end


--[[ ---------------------------------------------------------------------------
	 Scan debuffs if the target has had an aura event happen
----------------------------------------------------------------------------- ]]
function DebuffEnough:UNIT_AURA(_, unit)

	if (unit == "target") then
		self:ScanDebuffs()
	end

end


--[[ ---------------------------------------------------------------------------
	 Scans a hostile target to determine what debuffs are applied
----------------------------------------------------------------------------- ]]
function DebuffEnough:ScanDebuffs()

	self.tooltip = ""

	if (not UnitCanAttack("player", "target")) then
		self:UpdateVisible()
		return
	end
	
	-- reset applied flag for all tracked debuffs
	for spellName, _ in pairs(self.trackedDebuffs) do
		self.trackedDebuffs[spellName].applied = false
	end
	
	-- Check debuffs applied to target
	local i = 1
	while true do
	
		local debuff = UnitDebuff("target", i)
		if not debuff then break end
		
		if self.trackedDebuffs[debuff] then
			self.trackedDebuffs[debuff].applied = true
		end
		
		i = i + 1
	
	end
	
	self:RenderTooltip()
	
	self:UpdateVisible()
	
end


--[[ ---------------------------------------------------------------------------
	 Fill in the tooltip from scan results
----------------------------------------------------------------------------- ]]
function DebuffEnough:RenderTooltip()

	-- Determine best debuff for each slot
	-- {names = {}, slot = spellData.slot, applied = false}
	local processedDebuffs = {}
	local bestSlot = {}
	for spellName, spellData in pairs(self.trackedDebuffs) do
	
		if not spellData.applied then
	
			if not processedDebuffs[spellData.slot] then
				processedDebuffs[spellData.slot] = {}
			end
		
			if not processedDebuffs[spellData.slot][spellName] then
				processedDebuffs[spellData.slot][spellName] = {}
			end
		
			for name, debuffValue in pairs(spellData.names) do
				processedDebuffs[spellData.slot][spellName][name] = debuffValue
			end
			
		end
	
	end

	-- Render tooltip from processed results
	for debuffName, spells in pairs(processedDebuffs) do
	
		self.tooltip = self.tooltip.."\n\n"..debuffName
	
		for spellName, names in pairs(spells) do
		
			self.tooltip = self.tooltip.."\n  |cffffffff"..spellName.." ("
		
			for name, debuffValue in pairs(names) do
		
				local classColor = RAID_CLASS_COLORS[select(2, UnitClass(name))]
				local classColorStr = format(" |cff%02x%02x%02x%s|r", classColor.r*255, classColor.g*255, classColor.b*255, name)
				self.tooltip = self.tooltip..classColorStr
				
			end
			
			self.tooltip = self.tooltip.." |cffffffff)|r"
			
		end
		
	end
	
	-- Feed results to DO
	if not self.isDebuffEnough then
		self.dobj.text = L["No"]
	elseif self.isDebuffWarning then
		self.dobj.text = L["Warning"]
	else
		self.dobj.text = L["Yes"]
	end
	
	self.dobj.debuffenoughtooltiptext = self.tooltip
	
end